﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;
using DShowNET;

namespace LibCapture
{
    // デリゲート
    internal delegate void CapturedImageEventHandler( object sender, CapturedImageArgs args );

    public class CapturedImageArgs : EventArgs
    {
        public byte[] CapturedImage { get; private set; }
        public int DeviceIndex { get; private set; }
        public double SampleTime { get; private set; }

        public CapturedImageArgs( byte[] CapturedImage, int DeviceIndex )
            : this( CapturedImage, DeviceIndex, 0 )
        {
        }

        public CapturedImageArgs( byte[] CapturedImage, int DeviceIndex, double SampleTime )
        {
            this.CapturedImage = CapturedImage;
            this.DeviceIndex = DeviceIndex;
            this.SampleTime = SampleTime;
        }
    }

    internal class CaptureDevice : ISampleGrabberCB, IDisposable
    {
        private const int WS_CHILD = 0x40000000;		//VideoWindowの属性．
        private const int WS_CLIPCHILDREN = 0x02000000; //（Win32APIのCreateWindowExと同様のものを利用可能．）
        public const int WM_GRAPHNOTIFY = 0x00008001;	//DirectShowイベントの発生を示すWindows メッセージ.

        private readonly BackgroundWorker m_Worker = null;
        private readonly IntPtr m_PanelHandle = IntPtr.Zero;
        private readonly Rectangle m_PanelArea = Rectangle.Empty;
        private bool CaptureFlag = false;

        private IBaseFilter CaptureFilter { get; set; }
        private ISpecifyPropertyPages PropertyPages { get; set; }
        private IGraphBuilder GraphBuilder { get; set; }
        private ICaptureGraphBuilder2 CaptureGraphBuilder { get; set; }
        private ISampleGrabber SampleGrabber { get; set; }
        private IMediaControl MediaControl { get; set; }
        private IVideoWindow VideoWindow { get; set; }
        private IMediaEventEx MediaEvent { get; set; }
        private IBaseFilter GrabFilter { get; set; }
        private VideoInfoHeader VideoInfo { get; set; }

        public int DeviceIndex { get; private set; }
        public IWin32Window ParentWindow { get; private set; }
        public OnCaptureCompleted CapturedEvent { get; private set; }

        public CaptureDevice( int DeviceIndex, IWin32Window Window, IntPtr PanelHandle, Rectangle PanelArea,
            OnCaptureCompleted CapturedEvent )
        {
            // 
            this.DeviceIndex = DeviceIndex;
            ParentWindow = Window;
            m_PanelHandle = PanelHandle;
            m_PanelArea = PanelArea;
            this.CapturedEvent = CapturedEvent;

            // 
            m_Worker = new BackgroundWorker();
            m_Worker.WorkerReportsProgress = false;
            m_Worker.WorkerSupportsCancellation = false;
            m_Worker.DoWork += new DoWorkEventHandler( OnCapture );
            m_Worker.RunWorkerCompleted += new RunWorkerCompletedEventHandler( OnCaptureCompleted );
        }

        public bool InitCaptureDevice( UCOMIMoniker moniker )
        {
            try {
                // 
                if( !CreateCaptureDevice( moniker ) )
                    return false;

                // 
                if( !GetInterfaces() )
                    return false;

                // 
                if( !SetupGraph() )
                    return false;

                // 
                if( m_PanelHandle != IntPtr.Zero && !SetupVideoWindow() )
                    return false;

                // 
                int result = MediaControl.Run();
                if( result < 0 )
                    Marshal.ThrowExceptionForHR( result );

                return true;
            }
            catch( Exception ex ) {
                MessageBox.Show( ParentWindow, "ビデオストリームを正常に読み込めません。\n\n" + ex.Message,
                    "デバイスの初期化エラー", MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
            }

            return false;
        }

        private bool CreateCaptureDevice( UCOMIMoniker moniker )
        {
            object CaptureObject = null;

            try {
                // 
                Guid GuidBF = typeof( IBaseFilter ).GUID;
                moniker.BindToObject( null, null, ref GuidBF, out CaptureObject );
                CaptureFilter = (IBaseFilter)CaptureObject;
                CaptureObject = null;

                // 
                PropertyPages = (ISpecifyPropertyPages)CaptureFilter;

                return true;
            }
            catch( Exception ex ) {
                MessageBox.Show( ParentWindow, "キャプチャデバイスを作成できません。\n\n" + ex.Message,
                    "デバイスの初期化エラー", MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
                return false;
            }
            finally {
                if( CaptureObject != null ) {
                    // 
                    Marshal.ReleaseComObject( CaptureObject );
                }
            }
        }

        private bool GetInterfaces()
        {
            object ComObject = null;

            try {
                // 
                Type ComType = Type.GetTypeFromCLSID( Clsid.FilterGraph );
                if( ComType == null )
                    throw new ApplicationException( "FiterGraphオブジェクトが作成できません。" );
                ComObject = Activator.CreateInstance( ComType );
                GraphBuilder = (IGraphBuilder)ComObject;
                ComObject = null;

                // 
                ComType = Type.GetTypeFromCLSID( Clsid.CaptureGraphBuilder2 );
                if( ComType == null )
                    throw new ApplicationException( "CaptureGraphBuilder2オブジェクトが作成できません。" );
                ComObject = Activator.CreateInstance( ComType );
                CaptureGraphBuilder = (ICaptureGraphBuilder2)ComObject;
                ComObject = null;

                // 
                ComType = Type.GetTypeFromCLSID( Clsid.SampleGrabber );
                if( ComType == null )
                    throw new ApplicationException( "SampleGrabberオブジェクトが作成できません。" );
                ComObject = Activator.CreateInstance( ComType );
                SampleGrabber = (ISampleGrabber)ComObject;
                ComObject = null;

                // 
                MediaControl = (IMediaControl)GraphBuilder;
                VideoWindow = (IVideoWindow)GraphBuilder;
                MediaEvent = (IMediaEventEx)GraphBuilder;

                // 
                GrabFilter = (IBaseFilter)SampleGrabber;

                return true;
            }
            catch( Exception ex ) {
                MessageBox.Show( ParentWindow, "インターフェースの取得に失敗しました。\n\n" + ex.Message,
                    "デバイスの初期化エラー", MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
                return false;
            }
            finally {
                if( ComObject != null ) {
                    // 
                    Marshal.ReleaseComObject( ComObject );
                }
            }
        }

        private bool SetupGraph()
        {
            try {
                // 
                int result = CaptureGraphBuilder.SetFiltergraph( GraphBuilder );
                if( result < 0 ) {
                    Console.WriteLine( "SetFiltergraph : {0}", result );
                    Marshal.ThrowExceptionForHR( result );
                }

                // 
                result = GraphBuilder.AddFilter( CaptureFilter, "Video Capture Device" );
                if( result < 0 )
                    Marshal.ThrowExceptionForHR( result );

                // 
                AMMediaType ammedia = new AMMediaType();
                ammedia.majorType = MediaType.Video;
                ammedia.subType = MediaSubType.RGB24;
                ammedia.formatType = FormatType.VideoInfo;
                result = SampleGrabber.SetMediaType( ammedia );
                if( result < 0 )
                    Marshal.ThrowExceptionForHR( result );

                // 
                result = GraphBuilder.AddFilter( GrabFilter, "Frame Grab Filter" );
                if( result < 0 )
                    Marshal.ThrowExceptionForHR( result );

                // 
                Guid pinCategory = PinCategory.Capture;
                Guid media = MediaType.Video;
                result = CaptureGraphBuilder.RenderStream( ref pinCategory, ref media, CaptureFilter, null, GrabFilter );
                if( result < 0 )
                    Marshal.ThrowExceptionForHR( result );

                // パネル出力しない場合にはプレービューを組み込まない
                if( m_PanelHandle != IntPtr.Zero ) {
                    pinCategory = PinCategory.Preview;
                    media = MediaType.Video;
                    result = CaptureGraphBuilder.RenderStream( ref pinCategory, ref media, CaptureFilter, null, null );
                    if( result < 0 )
                        Marshal.ThrowExceptionForHR( result );
                }

                // 
                ammedia = new AMMediaType();
                result = SampleGrabber.GetConnectedMediaType( ammedia );
                if( result < 0 )
                    Marshal.ThrowExceptionForHR( result );
                if( (ammedia.formatType != FormatType.VideoInfo) || (ammedia.formatPtr == IntPtr.Zero) )
                    throw new ApplicationException( "キャプチャ出来ないメディアフォーマットです。" );

                // 
                VideoInfo = (VideoInfoHeader)Marshal.PtrToStructure( ammedia.formatPtr, typeof( VideoInfoHeader ) );
                Marshal.FreeCoTaskMem( ammedia.formatPtr );
                ammedia.formatPtr = IntPtr.Zero;

                //
                result = SampleGrabber.SetBufferSamples( false );
                if( result == 0 )
                    result = SampleGrabber.SetOneShot( false );
                if( result == 0 ) {
                    // パネル出力しない場合は、初期化と同時にサンプル開始
                    if( m_PanelHandle != IntPtr.Zero )
                        result = SampleGrabber.SetCallback( null, 0 );
                    else
                        result = SampleGrabber.SetCallback( this, 1 );
                }
                if( result < 0 )
                    Marshal.ThrowExceptionForHR( result );

                return true;
            }
            catch( Exception ex ) {
                MessageBox.Show( ParentWindow, "フィルターグラフの設定に失敗しました。\n\n" + ex.Message,
                    "デバイスの初期化エラー", MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
                return false;
            }
        }

        private bool SetupVideoWindow()
        {
            try {
                // 
                int result = VideoWindow.put_Owner( m_PanelHandle );
                if( result < 0 )
                    Marshal.ThrowExceptionForHR( result );

                // 
                result = VideoWindow.put_WindowStyle( WS_CHILD | WS_CLIPCHILDREN );
                if( result < 0 )
                    Marshal.ThrowExceptionForHR( result );

                // 
                VideoWindow.SetWindowPosition( 0, 0, m_PanelArea.Width, m_PanelArea.Height );

                // 
                result = VideoWindow.put_Visible( DsHlp.OATRUE );
                if( result < 0 )
                    Marshal.ThrowExceptionForHR( result );

                // 
                result = MediaEvent.SetNotifyWindow( ParentWindow.Handle, WM_GRAPHNOTIFY, IntPtr.Zero );
                if( result < 0 )
                    Marshal.ThrowExceptionForHR( result );

                return true;
            }
            catch( Exception ex ) {
                MessageBox.Show( ParentWindow, "ビデオウィンドウの初期化に失敗しました。\n\n" + ex.Message,
                    "デバイスの初期化エラー", MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
                return false;
            }
        }

        public void CaptureImage()
        {
            // 
            if( SampleGrabber == null )
                return;

            CaptureFlag = false;

            try {
                SampleGrabber.SetCallback( this, 1 );
            }
            catch( Exception ex ) {
                MessageBox.Show( ParentWindow, "フレームキャプチャに失敗しました。\n\n" + ex.Message,
                    "キャプチャエラー", MessageBoxButtons.OK, MessageBoxIcon.Information );
            }
        }

        public void OnGraphNotify()
        {
            // 
            if( MediaEvent != null )
                return;

            // 
            int result = 0;
            while( result == 0 ) {
                DsEvCode code;
                int param1, param2;
                result = MediaEvent.GetEvent( out code, out param1, out param2, 0 );
                if( result < 0 )
                    break;
                result = MediaEvent.FreeEventParams( code, param1, param2 );
            }
        }

        int ISampleGrabberCB.SampleCB( double SampleTime, IMediaSample pSample )
        {
            Console.WriteLine( "SampleCB" );
            return 0;
        }

        int ISampleGrabberCB.BufferCB( double SampleTime, IntPtr pBuffer, int BufferLength )
        {
            //Console.WriteLine( "BufferCB : {0}", Thread.CurrentThread.ManagedThreadId );
            // 
            if( m_PanelHandle != IntPtr.Zero && CaptureFlag )
                return 0;

            CaptureFlag = true;

            try {
                // 
                if( (pBuffer != IntPtr.Zero) && (BufferLength > 0) &&
                    (BufferLength <= VideoInfo.BmiHeader.ImageSize) ) {
                    int size = VideoInfo.BmiHeader.ImageSize;
                    var Frame = new byte[size];
                    Marshal.Copy( pBuffer, Frame, 0, BufferLength );

                    // 
                    if( !m_Worker.IsBusy )
                        m_Worker.RunWorkerAsync( new object[] { Frame, SampleTime } );
                }
                else
                    return 0;
            }
            catch( Exception ex ) {
                Console.WriteLine( "フレームキャプチャに失敗しました : {0}", ex.Message );
            }

            return 0;
        }

        private void OnCapture( object sender, DoWorkEventArgs e )
        {
            //Console.WriteLine( "OnCapture : {0}", Thread.CurrentThread.ManagedThreadId );
            var obj = e.Argument as object[];
            if( obj == null )
                return;
            var frame = obj[0] as byte[];
            if( frame == null )
                return;

            try {
                if( m_PanelHandle != IntPtr.Zero ) {
                    // サンプリングの停止
                    int result = SampleGrabber.SetCallback( null, 0 );
                }

                // 
                var data = new byte[frame.Length];
                Array.Copy( frame, data, frame.Length );

                // 
                e.Result = new object[] { data, obj[1] };
            }
            catch( Exception ex ) {
                Console.WriteLine( "OnCapture : {0}", ex.Message );
            }
        }

        private void OnCaptureCompleted( object sender, RunWorkerCompletedEventArgs e )
        {
            var obj = e.Result as object[];
            var bmp = obj[0] as byte[];
            if( bmp == null )
                return;
            var time = (double)obj[1];

            // 取得したビットマップイメージを通知
            if( m_PanelHandle != IntPtr.Zero )
                this.CapturedImageEvent( new CapturedImageArgs( e.Result as byte[], DeviceIndex ) );
            else {
                this.CapturedImageEvent( new CapturedImageArgs( bmp, DeviceIndex, time ) );
            }
        }

        public void Dispose()
        {
            CloseInterfaces();
        }

        public void CloseInterfaces()
        {
            try {
                if( MediaControl != null ) {
                    MediaControl.Stop();
                    Thread.Sleep( 100 );
                    MediaControl = null;
                }

                if( MediaEvent != null ) {
                    MediaEvent.SetNotifyWindow( IntPtr.Zero, WM_GRAPHNOTIFY, IntPtr.Zero );
                    MediaEvent = null;
                }

                if( VideoWindow != null ) {
                    VideoWindow.put_Visible( DsHlp.OAFALSE );
                    VideoWindow.put_Owner( IntPtr.Zero );
                    VideoWindow = null;
                }

                if( GrabFilter != null )
                    GrabFilter = null;

                if( SampleGrabber != null ) {
                    Marshal.ReleaseComObject( SampleGrabber );
                    SampleGrabber = null;
                }

                if( CaptureGraphBuilder != null ) {
                    Marshal.ReleaseComObject( CaptureGraphBuilder );
                    CaptureGraphBuilder = null;
                }

                if( GraphBuilder != null ) {
                    Marshal.ReleaseComObject( GraphBuilder );
                    GraphBuilder = null;
                }

                if( CaptureFilter != null ) {
                    CaptureFilter.Stop();
                    Marshal.ReleaseComObject( CaptureFilter );
                    CaptureFilter = null;
                }
            }
            catch( Exception ex ) {
                MessageBox.Show( ParentWindow, "インターフェースの終了に失敗しました。\n\n" + ex.Message,
                    "インターフェースエラー", MessageBoxButtons.OK, MessageBoxIcon.Information );
            }
        }

        public event CapturedImageEventHandler OnCapturedImage = null;
        private void CapturedImageEvent( CapturedImageArgs args )
        {
            if( this.OnCapturedImage != null ) {
                this.OnCapturedImage( this, args );
            }
        }
    }
}
