﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;
using DShowNET;
using DShowNET.Device;

namespace LibCapture
{
    // デリゲート
    public delegate void OnCaptureCompleted( byte[] image, int DeviceIndex, double SampleTime );

    public class MultiDevices
    {
        public const int WM_GRAPHNOTIFY = CaptureDevice.WM_GRAPHNOTIFY;	//DirectShowイベントの発生を示すWindows メッセージ.

        private List<CaptureDevice> m_Capture = new List<CaptureDevice>();
        private List<DsDevice> Devices = new List<DsDevice>();
        private List<string> DeviceNames = new List<string>();
        private List<bool> m_UseIndex = new List<bool>();
        private List<bool> m_IsCaptuer = new List<bool>();
        private List<bool> m_IsThrough = new List<bool>();
        private List<string> ClassID = new List<string>();

        public MultiDevices()
        {
            // 
            if( !DsUtils.IsCorrectDirectXVersion() ) {
                throw new ApplicationException( "アプリケーションの動作には、DirectX 8.1 以上が必要です。" );
            }

            // 
            ArrayList CaptureDevices;
            if( !DsDev.GetDevicesOfCat( FilterCategory.VideoInputDevice, out CaptureDevices ) ) {
                throw new ApplicationException( "ビデオキャプチャ可能なデバイスが見つかりません。" );
            }

            // 
            foreach( DsDevice dev in CaptureDevices ) {
                Devices.Add( dev );
                DeviceNames.Add( dev.Name );
                Guid id;
                dev.Mon.GetClassID( out id );
                ClassID.Add( id.ToString() );
            }

            try {
                // 
                foreach( string name in DeviceNames ) {
                    m_Capture.Add( null );
                    m_UseIndex.Add( false );
                    m_IsCaptuer.Add( false );
                    m_IsThrough.Add( false );
                }
            }
            catch( Exception ex ) {
                throw ex;
            }
        }

        public string[] EnumDeviceNames { get { return DeviceNames.ToArray(); } }
        public string[] EnumClassID { get { return ClassID.ToArray(); } }
        public int CountDevices { get { return Devices.Count; } }
        public bool IsCaptured
        {
            get
            {
                bool result = false;
                foreach( bool cap in m_IsCaptuer ) {
                    if( cap ) {
                        result = true;
                        break;
                    }
                }

                return result;
            }
        }

        public bool CreateCaptureDevice( int DeviceIndex, IWin32Window Parent, IntPtr PanelHandle, Rectangle PanelArea, bool CopyImage )
        {
            return CreateCaptureDevice( DeviceIndex, Parent, PanelHandle, PanelArea, CopyImage, null );
        }

        public bool CreateCaptureDevice( int DeviceIndex, IWin32Window Parent,
            IntPtr PanelHandle, Rectangle PanelArea, bool CopyImage, OnCaptureCompleted CaptureFunc )
        {
            // デバイスインデックスが範囲外または、デバイスが使用中の場合は処理を抜ける
            if( DeviceIndex >= CountDevices || m_UseIndex[DeviceIndex] )
                return false;

            // 使用中ではないデバイスが初期化されていない場合はデバイスを閉じる
            if( m_Capture[DeviceIndex] != null )
                m_Capture[DeviceIndex].CloseInterfaces();

            // キャプチャデバイスの作成
            m_Capture[DeviceIndex] = new CaptureDevice( DeviceIndex, Parent, PanelHandle, PanelArea, CaptureFunc );
            if( CaptureFunc != null ) {
                m_Capture[DeviceIndex].OnCapturedImage += new CapturedImageEventHandler( OnCapturedImage );
            }
            m_Capture[DeviceIndex].InitCaptureDevice( Devices[DeviceIndex].Mon );
            m_UseIndex[DeviceIndex] = true;
            if( PanelHandle == IntPtr.Zero )
                m_IsThrough[DeviceIndex] = true;

            return true;
        }

        public bool IsUsed( int DeviceIndex )
        {
            return m_UseIndex[DeviceIndex];
        }

        private void OnCapturedImage( object sender, CapturedImageArgs args )
        {
            //Console.WriteLine( "OnCapturedImage : {0}", Thread.CurrentThread.ManagedThreadId );

            try {
                // 
                if( !m_IsThrough[args.DeviceIndex] && !m_IsCaptuer[args.DeviceIndex] )
                    throw new ApplicationException( "画像取り込み中でないデバイスからの読取り完了通知" );
                m_IsCaptuer[args.DeviceIndex] = false;

                // 画像取り込みの完了
                if( m_Capture[args.DeviceIndex].ParentWindow == null )
                    return;
                else if( m_Capture[args.DeviceIndex].ParentWindow is Form ) {
                    Form parent = m_Capture[args.DeviceIndex].ParentWindow as Form;
                    parent.BeginInvoke(
                        m_Capture[args.DeviceIndex].CapturedEvent,
                        args.CapturedImage, args.DeviceIndex, args.SampleTime );
                }
                else if( m_Capture[args.DeviceIndex].ParentWindow is Control ) {
                    Control parent = m_Capture[args.DeviceIndex].ParentWindow as Control;
                    parent.BeginInvoke(
                        m_Capture[args.DeviceIndex].CapturedEvent,
                        args.CapturedImage, args.DeviceIndex, args.SampleTime );
                }
                else
                    throw new ApplicationException( "画像取り込み通知先が不明" );
            }
            catch( Exception ex ) {
                Console.WriteLine( "OnCapturedImage : {0}", ex.Message );
            }
        }

        public void CaptureImage( int DeviceIndex )
        {
            if( DeviceIndex < CountDevices && m_UseIndex[DeviceIndex] && !m_IsCaptuer[DeviceIndex] ) {
                m_Capture[DeviceIndex].CaptureImage();
                m_IsCaptuer[DeviceIndex] = true;
            }
        }

        public void CaptureImages()
        {
            for( int i = 0; i < CountDevices; i++ ) {
                if( m_UseIndex[i] && !m_IsCaptuer[i] ) {
                    m_Capture[i].CaptureImage();
                    m_IsCaptuer[i] = true;
                }
            }
        }

        public void OnGraphNotify()
        {
            for( int i = 0; i < CountDevices; i++ ) {
                if( m_UseIndex[i] )
                    m_Capture[i].OnGraphNotify();
            }
        }

        public void CloseDevice( int DeviceIndex )
        {
            if( DeviceIndex >= CountDevices || !m_UseIndex[DeviceIndex] )
                return;

            // 
            m_Capture[DeviceIndex].CloseInterfaces();
            m_UseIndex[DeviceIndex] = false;
        }

        public void CloseDevices()
        {
            for( int i = 0; i < m_Capture.Count; i++ ) {
                if( m_UseIndex[i] ) {
                    m_Capture[i].CloseInterfaces();
                    m_UseIndex[i] = false;
                }
            }

            // 
            //CaptureDevice.CloseDevices();
            //m_UseIndex.Clear();
        }
    }
}
