﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using LibCapture;

namespace MultiCapture
{
    public partial class Form1 : Form
    {
        private MultiDevices m_Capture = null;
        private readonly BackgroundWorker m_Worker0 = null;
        private readonly BackgroundWorker m_Worker1 = null;
        private readonly Font m_DisplayFont = new Font( "メイリオ", 10 );
        private double time0 = 0;
        private double time1 = 0;

        public Form1()
        {
            InitializeComponent();

            try {
                // 
                this.Frame0.Image = null;
                this.Frame1.Image = null;

                // 
                m_Worker0 = new BackgroundWorker();
                m_Worker0.WorkerReportsProgress = false;
                m_Worker0.WorkerSupportsCancellation = false;
                m_Worker0.DoWork += new DoWorkEventHandler( OnAnalize );
                m_Worker0.RunWorkerCompleted += new RunWorkerCompletedEventHandler( OnAnalizeCompleted );

                // 
                m_Worker1 = new BackgroundWorker();
                m_Worker1.WorkerReportsProgress = false;
                m_Worker1.WorkerSupportsCancellation = false;
                m_Worker1.DoWork += new DoWorkEventHandler( OnAnalize );
                m_Worker1.RunWorkerCompleted += new RunWorkerCompletedEventHandler( OnAnalizeCompleted );

                // キャプチャデバイスの作成
                m_Capture = new MultiDevices();

                // キャプチャデバイスの作成
                if( !m_Capture.CreateCaptureDevice( 0, this, IntPtr.Zero,
                    this.Frame0.ClientRectangle, true, CaptureCompleted ) ) {
                    MessageBox.Show( "起動エラー",
                        "起動エラー", MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
                    throw new ApplicationException( "起動エラー" );
                }
                if( !m_Capture.CreateCaptureDevice( 1, this, IntPtr.Zero,
                    this.Frame1.ClientRectangle, true, CaptureCompleted ) ) {
                    MessageBox.Show( "起動エラー",
                        "起動エラー", MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
                    throw new ApplicationException( "起動エラー" );
                }

                // デバイス名の表示
                this.Device0.Text = m_Capture.EnumDeviceNames[0];
                this.Device1.Text = m_Capture.EnumDeviceNames[1];
            }
            catch( Exception ex ) {
                MessageBox.Show( ex.Message,
                    "起動エラー", MessageBoxButtons.OK, MessageBoxIcon.Exclamation );
                this.Close();
            }

            // 
            this.FormClosed += new FormClosedEventHandler( Form1_FormClosed );
        }

        private void Form1_FormClosed( object sender, FormClosedEventArgs e )
        {
            // 注意：リソースの解放はもっと確実な方法で処理しましょう！


            if( m_Capture != null ) {
                // キャプチャデバイスの後片付け
                m_Capture.CloseDevices();
            }
        }

        protected override void WndProc( ref Message m )
        {
            if( m.Msg == MultiDevices.WM_GRAPHNOTIFY ) {
                m_Capture.OnGraphNotify();
                return;
            }

            base.WndProc( ref m );
        }

        private void CaptureCompleted( byte[] bitmap, int DeviceIndex, double SampleTime )
        {
            // 実行中の場合はフレームを破棄（フレーム落ちさせる）
            if( DeviceIndex == 0 && !m_Worker0.IsBusy )
                m_Worker0.RunWorkerAsync( new object[] { bitmap, DeviceIndex, SampleTime } );
            else if( DeviceIndex == 1 && !m_Worker1.IsBusy )
                m_Worker1.RunWorkerAsync( new object[] { bitmap, DeviceIndex, SampleTime } );
        }

        private void OnAnalize( object sender, DoWorkEventArgs e )
        {
            // 
            var obj = e.Argument as object[];
            var bitmap = obj[0] as byte[];
            var DeviceIndex = (int)obj[1];
            var SampleTime = (double)obj[2];

            try {
                // 
                Size size = DeviceIndex == 0 ? this.Frame0.Size : this.Frame1.Size;
                Bitmap dest = new Bitmap( size.Width, size.Height );
                using( Graphics graph = Graphics.FromImage( dest ) ) {
                    var sw = new Stopwatch();
                    sw.Start();

                    graph.InterpolationMode = InterpolationMode.NearestNeighbor;
                    Bitmap bmp;
                    try {
                        // webカメラの入力サイズによって変更する
                        int width = 640;
                        int height = 480;
                        int strige = width * 3;

                        // 
                        GCHandle gcHandle = GCHandle.Alloc( bitmap, GCHandleType.Pinned );

                        IntPtr addr = gcHandle.AddrOfPinnedObject();
                        addr = (IntPtr)(addr.ToInt32() + (height - 1) * strige);

                        // 画像データをビットマップに変換
                        bmp = new Bitmap( width, height, -strige, PixelFormat.Format24bppRgb, addr );
                        gcHandle.Free();

                        // ピクチャーボックス用に縮小処理
                        graph.DrawImage( bmp, 0, 0, size.Width, size.Height );
                    }
                    catch( Exception ex ) {
                        Console.WriteLine( "graph.DrawImage : {0}", DeviceIndex );
                        Console.WriteLine( ex.Message );
                        return;
                    }

                    // 
                    sw.Stop();
                    var time =
                        Stopwatch.IsHighResolution ?
                            (double)sw.ElapsedTicks / (double)Stopwatch.Frequency :
                            sw.Elapsed.TotalMilliseconds;


                    //---------------------------------------------------------
                    // ここに描画処理を記述
                    //---------------------------------------------------------


                    // 表示文字列の作成
                    var point = new PointF( 10, 5 );
                    var disp = string.Format( "{0:F1}FPS - {1:F6}",
                        1.0f / (double)(SampleTime - (DeviceIndex == 0 ? time0 : time1)), time );
                    // 縁取り
                    graph.DrawString( disp, m_DisplayFont, Brushes.White, new PointF( point.X - 1, 5 ) );
                    graph.DrawString( disp, m_DisplayFont, Brushes.White, new PointF( point.X + 1, 5 ) );
                    graph.DrawString( disp, m_DisplayFont, Brushes.White, new PointF( 10, point.Y - 1 ) );
                    graph.DrawString( disp, m_DisplayFont, Brushes.White, new PointF( 10, point.Y + 1 ) );
                    // 描画
                    graph.DrawString( disp, m_DisplayFont, Brushes.Red, point );

                    // 
                    if( DeviceIndex == 0 )
                        time0 = SampleTime;
                    else if( DeviceIndex == 1 )
                        time1 = SampleTime;

                    // 
                    bmp.Dispose();

                    // 
                    e.Result = new object[] { dest, DeviceIndex };
                }
            }
            catch( Exception ex ) {
                Console.WriteLine( "CaptureCompleted" );
                Console.WriteLine( ex.Message );
            }
        }

        private void OnAnalizeCompleted( object sender, RunWorkerCompletedEventArgs e )
        {
            // 
            if( e.Error != null )
                return;

            // 
            var obj = e.Result as object[];
            if( obj == null )
                return;
            var bitmap = obj[0] as Bitmap;
            var DeviceIndex = (int)obj[1];

            // 
            if( DeviceIndex == 0 ) {
                if( this.Frame0.Image != null )
                    this.Frame0.Image.Dispose();
                this.Frame0.Image = bitmap;
            }
            else if( DeviceIndex == 1 ) {
                if( this.Frame1.Image != null )
                    this.Frame1.Image.Dispose();
                this.Frame1.Image = bitmap;
            }
        }
    }
}
